# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
import json
from pathlib import Path

from ai4rag import logger

from .base_event_handler import BaseEventHandler, LogLevel


class LocalEventHandler(BaseEventHandler):
    """Implementation used for local development and testing purposes.

    Parameters
    ----------
    output_path : str | Path | None, default=None
        Location where results for all patterns will be saved.
    """

    def __init__(self, output_path: str | Path | None = None):
        self.output_path = Path(output_path) if output_path else None

    def on_status_change(self, level: LogLevel, message: str, step: str | None = None) -> None:
        logger.info("LocalEventHandler ::: %s ::: %s ::: %s", level, step, message)

    def on_pattern_creation(self, payload: dict, evaluation_results: list, **kwargs) -> None:
        logger.info("LocalEventHandler ::: Pattern creation ::: ", payload)
        pattern_name = payload.get("rag_pattern", {}).get("name", "default_pattern_name")

        if self.output_path:
            dir_path = self.output_path / pattern_name
            dir_path.mkdir(exist_ok=False, parents=True)

            evaluation_results_path = dir_path / "evaluation_results.json"
            with open(evaluation_results_path, mode="w") as file:
                json.dump(evaluation_results, file)

            with open(dir_path / "pattern.json", mode="w") as file2:
                json.dump(payload, file2)



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import Optional, TypedDict

from httpx import Timeout
from llama_stack_client import Client

from .base_model import BaseEmbeddingModel

__all__ = ["LSEmbeddingModel", "LSEmbeddingParams"]


class LSEmbeddingParams(TypedDict, total=False):
    """LLamaStack parameters to be used to create embeddings."""

    embedding_dimension: int
    context_length: int
    timeout: Optional[float | Timeout]
    model_type: Optional[str]
    provider_id: Optional[str]
    provider_resource_id: Optional[str]


class LSEmbeddingModel(BaseEmbeddingModel[Client, LSEmbeddingParams]):
    """Creates embeddings for LLamaStack client."""

    def __init__(self, client: Client, model_id: str, params: LSEmbeddingParams):
        params = params or {}
        super().__init__(client=client, model_id=model_id, params=params)

    def _embed_text(self, text_input: list[str] | str) -> list[list[float]]:
        """Embeds documents.

        Parameters
        ----------
        text_input : list[str] | str
            List of text-like chunks or single text-like chunk.

        Returns
        -------
        list[list[float]]
            Embeddings made from the list of texts or a single text.
        """

        return [
            data.embedding
            for data in self.client.embeddings.create(input=text_input, model=self.model_id).data
            if not isinstance(data.embedding, str)
        ]

    def embed_documents(self, texts: list[str]) -> list[list[float]]:
        """Embeds given list of strings.

        Parameters
        ----------
        texts : list[str]
            List of text-like chunks.

        Returns
        -------
        list[list[float]]
            Embeddings made from the list of texts.
        """
        return self._embed_text(text_input=texts)

    def embed_query(self, query: str) -> list[float]:
        """Embeds given query.

        Parameters
        ----------
        query : str
            Single text-like chunk.

        Returns
        -------
        list[]
            Embeddings made from a single text.
        """
        return self._embed_text(text_input=query)[0]



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import Annotated, Optional

from annotated_types import Ge, Gt, Le, MinLen
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
)

from ai4rag.utils.constants import (
    ChatGenerationConstants,
)

config = ConfigDict(extra="forbid")


class AI4RAGFoundationModelParams(BaseModel):
    """Attributes to be included in the generation.foundation_models payload."""

    model_config = config

    max_completion_tokens: Annotated[int, Gt(0)] = ChatGenerationConstants.MAX_COMPLETION_TOKENS
    temperature: Annotated[float, Ge(0), Le(1)] = ChatGenerationConstants.TEMPERATURE


class AI4RAGFoundationModel(BaseModel):
    """Attributes to be included in the generation.foundation_models payload."""

    model_config = config

    model_id: Annotated[str, MinLen(1)]
    parameters: Optional[AI4RAGFoundationModelParams] = Field(default_factory=AI4RAGFoundationModelParams)


class AI4RAGConstraints(BaseModel):
    """Attributes to be included in constraints payload."""

    model_config = config

    embedding_models: Optional[Annotated[list[AI4RAGFoundationModel], MinLen(1)]] = None
    foundation_models: Optional[Annotated[list[AI4RAGFoundationModel], MinLen(1)]] = None



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
import pandas as pd


def handle_missing_values_in_combinations_being_explored(df: pd.DataFrame):
    """
    Each record in the dataframe created to fit the model may differ.
    For single configuration one feature may not exist when other features takes specific value
    (i.e. for different retrieval types settings may differ).

    This function exists to handle dataframes for model training despite such differentiation.

    Params
    ------
    df : pd.DataFrame
        Experiment data, i.e. combinations being explored throughout the experiment.

    Returns
    -------
    pd.DataFrame
        Experiment data with `NaN` values properly replaced by more meaningful data.

    Notes
    -----
    This is basically only needed for fitting and transforming data using sklearn encoders.
    """
    if "ranker_strategy" in df.columns:
        df["ranker_strategy"] = df["ranker_strategy"].map(lambda el: "" if pd.isna(el) else el)
        df["ranker_sparse_vectors"] = (
            df["ranker_sparse_vectors"]
            .map(lambda el: str(el) if isinstance(el, list) else el)
            .map(lambda el: "" if pd.isna(el) else el)
        )

    for hybrid_numerical_col in ("ranker_k", "ranker_alpha"):
        if hybrid_numerical_col in df.columns:
            df[hybrid_numerical_col] = df[hybrid_numerical_col].map(lambda el: -1 if pd.isna(el) else el)

    return df



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import Any, TypedDict

from llama_stack_client import LlamaStackClient
from pydantic import TypeAdapter, ValidationError

from ai4rag import logger
from ai4rag.rag.embedding.llama_stack import LSEmbeddingModel
from ai4rag.rag.foundation_models.llama_stack import LSFoundationModel
from ai4rag.search_space.prepare.input_payload_types import AI4RAGConstraints
from ai4rag.search_space.prepare.validation_error_decoder import validation_error_decoder
from ai4rag.search_space.src.exceptions import SearchSpaceValueError
from ai4rag.search_space.src.parameter import Parameter
from ai4rag.search_space.src.search_space import AI4RAGSearchSpace

__all__ = ["prepare_search_space_with_llama_stack"]


class _DefaultModelsResponseType(TypedDict):
    foundation_models: list[LSFoundationModel]
    embedding_models: list[LSEmbeddingModel]


def _get_default_llama_stack_models(client: LlamaStackClient) -> _DefaultModelsResponseType:
    """Get list of default foundation models based on the available ones in llama stack."""

    logger.info("Selecting default foundation models...")
    available_models = client.models.list()
    llms = [model for model in available_models if model.custom_metadata.get("model_type") == "llm"]
    embeddings = [model for model in available_models if model.custom_metadata.get("model_type") == "embedding"]
    foundation_models = [LSFoundationModel(model_id=m.id, client=client) for m in llms]
    embedding_models = [
        LSEmbeddingModel(
            model_id=m.id, client=client, params={"embedding_dimension": m.custom_metadata["embedding_dimension"]}
        )
        for m in embeddings
    ]

    if not foundation_models:
        raise SearchSpaceValueError("There are no available models of type 'llm'.")
    if not embedding_models:
        raise SearchSpaceValueError("There are no available models of type 'embedding'.")

    logger.info("Selected default foundation models: %s.", foundation_models)
    logger.info("Selected default embedding models: %s.", embedding_models)

    return {"foundation_models": foundation_models, "embedding_models": embedding_models}


def _are_provided_models_available(
    provided_models: list, available_models: list[LSFoundationModel | LSEmbeddingModel]
) -> bool:
    """Check whether models provided by the user are available for the experiment."""

    available_ids = [m.model_id for m in available_models]

    for model in provided_models:
        m_id = model.model_id
        if m_id not in available_ids:
            raise SearchSpaceValueError(f"Provided model with model_id: {m_id} is not available for the experiment.")
    return True


def prepare_search_space_with_llama_stack(payload: dict[str, Any], client: LlamaStackClient) -> AI4RAGSearchSpace:
    """
    Prepare AutoRAGSearchSpace.

    Parameters
    ----------
    payload : dict[str, Any]
        A mapping between parameter name and its associated values.

    client : LlamaStackClient
        Client instance for listing and validating available models.

    Returns
    -------
    AI4RAGSearchSpace
        A valid AI4RAGSearchSpace used in RAG optimization process.

    Raises
    ------
    SearchSpaceValueError
        Raised when payload contains non-recognized parameter name.
    """
    logger.info("Preparing search space based on provided constraints: %s.", payload)

    payload_model = TypeAdapter(AI4RAGConstraints)

    try:
        validated_payload = payload_model.validate_python(payload)
    except ValidationError as ve:
        # we want to catch only the first error
        validation_error_decoder(ve.errors()[0])

    if isinstance(client, LlamaStackClient):
        models = _get_default_llama_stack_models(client)
        default_foundation_models = models["foundation_models"]
        default_embedding_models = models["embedding_models"]
    else:
        raise SearchSpaceValueError(f"Unrecognized client type: {client.__class__.__name__}")

    if validated_payload.foundation_models:
        _are_provided_models_available(validated_payload.foundation_models, default_foundation_models)
    if validated_payload.embedding_models:
        _are_provided_models_available(validated_payload.embedding_models, default_embedding_models)

    # Transform user models into llama-stack based models
    if validated_payload.foundation_models is not None:
        fms_param = Parameter(
            name="foundation_model",
            param_type="C",
            values=[
                LSFoundationModel(
                    model_id=fm.model_id,
                    client=client,
                    model_params=fm.parameters.model_dump() if fm.parameters else {},
                )
                for fm in validated_payload.foundation_models
            ],
        )
    else:
        fms_param = Parameter(
            name="foundation_model",
            param_type="C",
            values=default_foundation_models,
        )

    if validated_payload.embedding_models is not None:
        ems_param = Parameter(
            name="embedding_model",
            param_type="C",
            values=[
                LSEmbeddingModel(
                    model_id=em.model_id, client=client, params=em.parameters.model_dump() if em.parameters else {}
                )
                for em in validated_payload.embedding_models
            ],
        )
    else:
        ems_param = Parameter(
            name="embedding_model",
            param_type="C",
            values=default_embedding_models,
        )

    return AI4RAGSearchSpace(
        params=[fms_param, ems_param],
    )



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import Literal

from langchain_core.documents import Document

from ai4rag.rag.vector_store.base_vector_store import BaseVectorStore


class Retriever:
    """Class responsible for retrieving data from given vector store.

    Parameters
    ----------
    vector_store : BaseVectorStore
        Vector store / vector index to retrieve data from.

    method : Literal["simple"]
        Method describing how data should be retrieved.

    number_of_chunks : int
        Number of chunks to retrieve.
    """

    def __init__(
        self,
        vector_store: BaseVectorStore,
        method: Literal["simple"],
        number_of_chunks: int,
    ):
        self._vector_store = vector_store
        self.method = method
        self.number_of_chunks = number_of_chunks

    def retrieve(self, query: str, **kwargs) -> list[dict]:
        """Retrieve relevant documents from vector store.

        Parameters
        ----------
        query : str
            question for which documents should be retrieved.

        Returns
        -------
        list[Document]
            list of documents with their metadata corresponding to the query.
        """
        _number_of_chunks = kwargs.get("number_of_chunks", self.number_of_chunks)

        return self._vector_store.search(query, k=_number_of_chunks)



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from abc import ABC, abstractmethod
from enum import StrEnum

__all__ = ["BaseEventHandler", "LogLevel"]


class LogLevel(StrEnum):
    """Available log levels."""

    INFO = "info"
    WARNING = "warning"
    ERROR = "error"


class BaseEventHandler(ABC):
    """
    Abstract class defining interface for streaming results and messages,
    to the service layer.
    """

    @abstractmethod
    def on_status_change(self, level: LogLevel, message: str, step: str | None = None) -> None:
        """
        Method called to notify about experiment's status change.

        Parameters
        ----------
        level : LogLevel
            Logging level

        message : str
            Text of streamed message

        step : str
            Currently performed step. It should be one of composition steps.
        """

    @abstractmethod
    def on_pattern_creation(self, payload: dict, evaluation_results: list, **kwargs) -> None:
        """
        Method called when single RAG pattern's evaluation is completed.

        Parameters
        ----------
        payload : dict
            Information about RAG pattern's location and name, calculated scores
            and message.

            Example:

            name --> Pattern 1
            iteration --> which pattern we are evaluating

            payload = {
                "metrics": {
                    "test_data": {
                        "answer_correctness": {
                            "mean": 0.7,
                            "ci_low": 0.6,
                            "ci_high": 0.8,
                        },
                    },
                },
                "context": {
                    "rag_pattern": {
                        "composition_steps" : [
                            "chunking", "embeddings", "vector_store", "retrieval", "generation"
                        ],
                        "duration": 3507.9,
                        "location": {},
                        "name": "Pattern 1",
                        "settings": {
                            "chunking": {
                                "method": "recursive",
                                "chunk_size": 256,
                                "overlap": 128
                            },
                            "embeddings": {
                                "truncate_strategy": "left",
                                "input_size": 384,
                                "model_name": "ibm/slate.30m.english.rtrvr"
                            },
                            "vector_store": {
                                "database": "milvus",
                                "index_name": "XD_1234_index_5678",
                                "distance_metric": "euclidean"
                                "operation": "upsert",
                                "document_schema": {...},
                            },
                            "retrieval": {
                                "method": "simple",
                                "number_of_chunks": 5
                            },
                            "generation": {
                                "model_name": "mistralai/mixtral-8x7b-instruct-v0-1",
                                "parameters": {
                                    "max_new_tokens": 256
                                },
                                "chat_template_messages": {
                                    "system_message_text": "...",
                                    "user_message_text": "...",
                                }
                                "context_template_text": "...",
                            }
                        }
                    }
                    "iteration": 1,
                    "max_combinations": 100
                }
            }

        evaluation_results : dict
            Results from single pattern evaluation.

            Example content:
            "data": [
                {
                    "question_id": "0",
                    "answer": "<model's answer>",
                    "answer_contexts": [
                        {"text": "<content1_text>", "document_id": "document_1.pdf"},
                        {"text": "<content2_text>", "document_id": "document_2.pdf"},
                    ]
                    "scores": {
                        "answer_correctness": 0.79,
                        "context_correctness": 0.65,
                    }
                },
                {
                    "question_id": "1",
                    "answer": "<model's answer>",
                    "answer_contexts": [
                        {"text": "<content3_text>", "document_id": "document_3.pdf"},
                        {"text": "<content4_text>", "document_id": "document_4.pdf"},
                    ]
                    "scores": {
                        "answer_correctness": 0.79,
                        "context_correctness": 0.65,
                    }
                }
            ]
        """



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from ai4rag.search_space.src.parameter import Parameter
from ai4rag.utils.constants import AI4RAGParamNames

__all__ = [
    "get_default_ai4rag_search_space_parameters",
]

_default_chunking_methods = ("recursive",)
_default_chunk_sizes = (1024, 2048)
_default_chunk_overlaps = (256, 512)
_default_retrieval_methods = ("simple",)
_default_window_sizes = (0,)
_default_numbers_of_chunks = (3, 5, 10)


def get_default_ai4rag_search_space_parameters() -> list[Parameter]:
    """
    Function to return default search space containing experiment parameters.

    Returns
    -------
    list[Parameter]
        Parameters that will be used for creating AI4RAGSearchSpace
    """

    default_search_space_parameters = [
        Parameter(
            name=AI4RAGParamNames.CHUNKING_METHOD,
            param_type="C",
            values=_default_chunking_methods,
        ),
        Parameter(
            name=AI4RAGParamNames.CHUNK_SIZE,
            param_type="C",
            values=_default_chunk_sizes,
        ),
        Parameter(
            name=AI4RAGParamNames.CHUNK_OVERLAP,
            param_type="C",
            values=_default_chunk_overlaps,
        ),
        Parameter(
            name=AI4RAGParamNames.RETRIEVAL_METHOD,
            param_type="C",
            values=_default_retrieval_methods,
        ),
        Parameter(name=AI4RAGParamNames.WINDOW_SIZE, param_type="C", values=_default_window_sizes),
        Parameter(
            name=AI4RAGParamNames.NUMBER_OF_CHUNKS,
            param_type="C",
            values=_default_numbers_of_chunks,
        ),
    ]

    return default_search_space_parameters



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import Any

from openai import OpenAI

from ai4rag.rag.embedding.base_model import BaseEmbeddingModel


class OpenAIEmbeddingModel(BaseEmbeddingModel):
    """Class wrapping independently deployed embedding model with OpenAI client."""

    def __init__(self, client: OpenAI, model_id: str, params: dict[str, Any] | None = None):
        super().__init__(client=client, model_id=model_id, params=params)

    def embed_documents(self, texts: list[str]) -> list[list[float]]:
        """Embeds given list of strings.

        Parameters
        ----------
        texts : list[str]
            List of text-like chunks.

        Returns
        -------
        list[list[float]]
            Embeddings made from the list of texts.
        """
        resp = self.client.embeddings.create(model=self.model_id, input=texts)
        return [d.embedding for d in resp.data]

    def embed_query(self, query: str) -> list[float]:
        """Embeds given query.

        Parameters
        ----------
        query : str
            Single text-like chunk.

        Returns
        -------
        list[]
            Embeddings made from a single text.
        """
        return self.client.embeddings.create(model=self.model_id, input=query).data[0].embedding



# -----------------------------------------------------------------------------
# Copyright IBM Corp. 2025-2026
# SPDX-License-Identifier: Apache-2.0
# -----------------------------------------------------------------------------
from typing import Annotated, Any

from annotated_types import Ge, Gt, Le
from llama_stack_client import LlamaStackClient
from pydantic import BaseModel

from ai4rag.rag.foundation_models.base_model import BaseFoundationModel
from ai4rag.utils.constants import ChatGenerationConstants


class ModelParameters(BaseModel):
    max_completion_tokens: Annotated[int, Gt(0)] = ChatGenerationConstants.MAX_COMPLETION_TOKENS
    temperature: Annotated[float, Ge(0), Le(1)] = ChatGenerationConstants.TEMPERATURE


class LSFoundationModel(BaseFoundationModel[LlamaStackClient, dict[str, Any] | ModelParameters | None]):
    """Integration point to use any model via Llama-stack API / client"""

    def __init__(
        self,
        client: LlamaStackClient,
        model_id: str,
        model_params: dict[str, Any] | ModelParameters | None = None,
        system_message_text: str | None = None,
        user_message_text: str | None = None,
        context_template_text: str | None = None,
    ):
        super().__init__(
            client=client,
            model_id=model_id,
            model_params=model_params,
            system_message_text=system_message_text,
            user_message_text=user_message_text,
            context_template_text=context_template_text,
        )

    def chat(self, system_message: str, user_message: str) -> str:
        """
        Chat completion for communication with selected foundation model.

        Parameters
        ----------
        system_message : str
            System messages in the str format.

        user_message : str
            User message in the str format.

        Returns
        -------
        str
            Chat response from the model.
        """
        response_chat = self.client.chat.completions.create(
            model=self.model_id,
            messages=[
                {"role": "system", "content": system_message},
                {"role": "user", "content": user_message},
            ],
        )
        answer = response_chat.choices[0].message.content

        return answer
